@@ -1,4 +1,4 @@
-# Copyright (c) 2022 CNES
+# Copyright (c) 2023 CNES
 #
 # All rights reserved. Use of this source code is governed by a
 # BSD-style license that can be found in the LICENSE file.
@@ -7,8 +7,17 @@ Handle a collection of Zarr groups.
 ===================================
 """
 from . import merging, partitioning
-from .collection import Collection, Indexer, PartitionFilter
-from .collection.callable_objects import PartitionCallable, UpdateCallable
+from .collection import (
+    Collection,
+    Indexer,
+    PartitionFilter,
+    PartitionFilterCallback,
+)
+from .collection.callable_objects import (
+    MapCallable,
+    PartitionCallable,
+    UpdateCallable,
+)
 from .convenience import (
     create_collection,
     create_view,
@@ -29,11 +38,13 @@ __all__ = [
     'create_view',
     'Dataset',
     'Indexer',
+    'MapCallable',
     'merging',
     'open_collection',
     'open_view',
     'PartitionCallable',
     'PartitionFilter',
+    'PartitionFilterCallback',
     'partitioning',
     'UpdateCallable',
     'Variable',
