@@ -1,4 +1,4 @@
-# Copyright (c) 2022 CNES
+# Copyright (c) 2023 CNES
 #
 # All rights reserved. Use of this source code is governed by a
 # BSD-style license that can be found in the LICENSE file.
@@ -21,10 +21,12 @@ from typing import (
     Union,
 )
 import datetime
+import functools
 import io
 import itertools
 import json
 import logging
+import os
 import pathlib
 import types
 
@@ -38,14 +40,15 @@ import zarr
 import zarr.storage
 
 from .. import (
+    dask_utils,
     dataset,
     expression,
+    fs_utils,
     merging,
     meta,
     partitioning,
     storage,
     sync,
-    utilities,
 )
 from .callable_objects import MapCallable, PartitionCallable, UpdateCallable
 from .detail import (
@@ -135,14 +138,12 @@ def build_indexer_args(
 
 
 def _immutable_path(
-    fs: fsspec.AbstractFileSystem,
     ds: meta.Dataset,
     partition_properties: PartitioningProperties,
 ) -> str | None:
     """Return the immutable path of the dataset.
 
     Args:
-        fs: The file system that the partition is stored on.
         ds: The dataset to process.
         partition_properties: The partitioning properties.
 
@@ -151,8 +152,8 @@ def _immutable_path(
         relative to the partitioning or None if the dataset does not contain
         immutable data.
     """
-    return fs.sep.join(
-        (partition_properties.dir, _IMMUTABLE)) if ds.select_variables_by_dims(
+    return fs_utils.join_path(
+        partition_properties.dir, _IMMUTABLE) if ds.select_variables_by_dims(
             (partition_properties.dim, ), predicate=False) else None
 
 
@@ -243,13 +244,12 @@ class Collection:
         #: The metadata that describes the dataset handled by the collection.
         self.metadata = ds
         #: The file system used to read/write the collection.
-        self.fs = utilities.get_fs(filesystem)
+        self.fs = fs_utils.get_fs(filesystem)
         #: The partitioning strategy used to split the data.
         self.partitioning = partition_handler
         #: The partitioning properties (base directory and dimension).
         self.partition_properties = PartitioningProperties(
-            utilities.normalize_path(self.fs,
-                                     partition_base_dir.rstrip(self.fs.sep)),
+            fs_utils.normalize_path(self.fs, partition_base_dir),
             ds.variables[axis].dimensions[0],
         )
         #: The access mode of the collection.
@@ -258,11 +258,7 @@ class Collection:
         self.synchronizer = synchronizer or sync.NoSync()
         #: The path to the dataset that contains the immutable data relative
         #: to the partitioning.
-        self._immutable = _immutable_path(
-            self.fs,
-            ds,
-            self.partition_properties,
-        )
+        self._immutable = _immutable_path(ds, self.partition_properties)
 
         self._write_config(skip_if_exists=True)
 
@@ -301,15 +297,14 @@ class Collection:
         raise io.UnsupportedOperation('not writable')
 
     @classmethod
-    def _config(cls, partition_base_dir: str,
-                fs: fsspec.AbstractFileSystem) -> str:
+    def _config(cls, partition_base_dir: str) -> str:
         """Return the configuration path."""
-        return fs.sep.join((partition_base_dir, cls.CONFIG))
+        return fs_utils.join_path(partition_base_dir, cls.CONFIG)
 
     def _write_config(self, skip_if_exists: bool = False) -> None:
         """Write the configuration file."""
         base_dir = self.partition_properties.dir
-        config = self._config(base_dir, self.fs)
+        config = self._config(base_dir)
         exists = self.fs.exists(config)
 
         if skip_if_exists and exists:
@@ -321,9 +316,11 @@ class Collection:
 
         self.fs.makedirs(base_dir, exist_ok=True)
 
-        params = dict(axis=self.axis,
-                      dataset=self.metadata.get_config(),
-                      partitioning=self.partitioning.get_config())
+        params = {
+            'axis': self.axis,
+            'dataset': self.metadata.get_config(),
+            'partitioning': self.partitioning.get_config(),
+        }
 
         with self.fs.open(config, mode='w') as stream:
             json.dump(params, stream, indent=4)  # type: ignore[arg-type]
@@ -356,8 +353,8 @@ class Collection:
             ValueError: If the provided directory does not contain a collection.
         """
         _LOGGER.info('Opening collection: %r', path)
-        fs = utilities.get_fs(filesystem)
-        config = cls._config(path, fs)
+        fs = fs_utils.get_fs(filesystem)
+        config = cls._config(path)
         if not fs.exists(config):
             raise ValueError(f'zarr collection not found at path {path!r}')
         with fs.open(config) as stream:
@@ -407,8 +404,8 @@ class Collection:
                 None, the new partitioned data overwrites the existing
                 partitioned data.
             npartitions: The maximum number of partitions to process in
-                parallel. By default, the number of available cores of the dask
-                cluster is used.
+                parallel. By default, all partitions are processed in parallel
+                in a single task.
 
         .. warning::
 
@@ -434,7 +431,7 @@ class Collection:
             variable = self.metadata.variables[item]
             ds.add_variable(variable)
 
-        client = utilities.get_client()
+        client = dask_utils.get_client()
 
         # If the dataset contains variables that should not be partitioned.
         if self._immutable is not None:
@@ -448,15 +445,29 @@ class Collection:
             # Remove the variables that should not be partitioned.
             ds = ds.select_variables_by_dims((self.axis, ))
 
-        utilities.calculation_stream(
-            _insert,
-            self.partitioning.split_dataset(ds, self.partition_properties.dim),
-            max_workers=npartitions,
-            axis=self.axis,
-            ds=client.scatter(ds),
-            fs=self.fs,
-            merge_callable=merge_callable,
-            partitioning_properties=self.partition_properties)
+        # Process the partitions to insert or update by batches to avoid
+        # memory issues.
+        partitions = tuple(
+            self.partitioning.split_dataset(ds, self.partition_properties.dim))
+
+        if npartitions is not None:
+            if npartitions < 1:
+                raise ValueError('The number of partitions must be positive')
+            npartitions = len(partitions) // npartitions + 1
+
+        scattared_ds = client.scatter(ds)
+        for sequence in dask_utils.split_sequence(partitions, npartitions):
+            futures = [
+                dask_utils.simple_delayed('insert', _insert)(
+                    partition,  # type: ignore[arg-type]
+                    axis=self.axis,
+                    ds=scattared_ds,
+                    fs=self.fs,
+                    merge_callable=merge_callable,
+                    partitioning_properties=self.partition_properties,
+                ) for partition in sequence
+            ]
+            client.compute(futures, sync=True)
 
     def _relative_path(self, path: str) -> str:
         """Return the relative path to the collection.
@@ -540,7 +551,7 @@ class Collection:
             ...     timedelta=datetime.timedelta(days=30))
         """
         now = datetime.datetime.now()
-        client = utilities.get_client()
+        client = dask_utils.get_client()
         folders = list(self.partitions(filters=filters))
 
         def is_created_before(path: str) -> bool:
@@ -708,12 +719,13 @@ class Collection:
                 The result of the function.
             """
             ds, indices = _load_dataset_with_overlap(
-                self.axis, depth, self.partition_properties.dim, self.fs,
-                self._immutable, partition, partitions, selected_variables)
+                depth, self.partition_properties.dim, self.fs, self._immutable,
+                partition, partitions, selected_variables)
 
             if add_partition_info:
                 kwargs = kwargs.copy()
-                kwargs['partition_info'] = indices
+                kwargs['partition_info'] = (self.partition_properties.dim,
+                                            indices)
 
             # Finally, apply the function.
             return (self.partitioning.parse(partition),
@@ -760,7 +772,7 @@ class Collection:
             ...     filters=lambda keys: keys["year"] == 2019 and
             ...     keys["month"] == 3 and keys["day"] % 2 == 0)
         """
-        client = utilities.get_client()
+        client = dask_utils.get_client()
         arrays: list[dataset.Dataset] = []
         if indexer is None:
             selected_partitions = tuple(self.partitions(filters=filters))
@@ -771,7 +783,7 @@ class Collection:
             # selected partition.
             bag = dask.bag.core.from_sequence(
                 self.partitions(filters=filters),
-                npartitions=utilities.dask_workers(client, cores_only=True))
+                npartitions=dask_utils.dask_workers(client, cores_only=True))
             arrays = bag.map(storage.open_zarr_group,
                              fs=self.fs,
                              selected_variables=selected_variables).compute()
@@ -783,7 +795,7 @@ class Collection:
 
             bag = dask.bag.core.from_sequence(
                 args,
-                npartitions=utilities.dask_workers(client, cores_only=True))
+                npartitions=dask_utils.dask_workers(client, cores_only=True))
 
             # Finally, load the selected partitions and apply the indexer.
             arrays = list(
@@ -869,15 +881,15 @@ class Collection:
                                            selected_variables, *args, **kwargs)
         else:
             local_func = _wrap_update_func_with_overlap(
-                self.axis, depth, self.partition_properties.dim, func, self.fs,
+                depth, self.partition_properties.dim, func, self.fs,
                 self._immutable, selected_variables, *args, **kwargs)
 
         _LOGGER.info('Updating of the (%s) variable in the collection',
                      ', '.join(repr(item) for item in func_result))
-        client = utilities.get_client()
-        batchs = utilities.split_sequence(
+        client = dask_utils.get_client()
+        batchs = dask_utils.split_sequence(
             tuple(self.partitions(filters=filters)), partition_size
-            or utilities.dask_workers(client, cores_only=True))
+            or dask_utils.dask_workers(client, cores_only=True))
         awaitables = client.map(local_func, tuple(batchs), key=func.__name__)
         storage.execute_transaction(client, self.synchronizer, awaitables)
 
@@ -993,9 +1005,12 @@ class Collection:
             The iterator over the partitions and the zarr groups.
         """
         yield from (
-            (self._relative_path(item) if relative else item,
-             zarr.open_consolidated(self.fs.get_mapper(item)))  # type: ignore
-            for item in self.partitions())
+            (
+                self._relative_path(item) if relative else item,
+                zarr.open_consolidated(
+                    self.fs.get_mapper(item),  # type: ignore
+                    mode='r',
+                )) for item in self.partitions())
 
     def variables(
         self,
@@ -1011,3 +1026,72 @@ class Collection:
             The variables of the collection.
         """
         return variables(self.metadata, selected_variables)
+
+    def copy(
+        self,
+        target: str,
+        *,
+        filters: PartitionFilter | None = None,
+        filesystem: fsspec.AbstractFileSystem | None = None,
+        mode: str = 'w',
+        npartitions: int | None = None,
+        synchronizer: sync.Sync | None = None,
+    ) -> Collection:
+        """Copy the collection to a new location.
+
+        Args:
+            target: The target location.
+            filters: The predicate used to filter the partitions to copy.
+            filesystem: The file system to use. If None, the file system of the
+                collection is used.
+            mode: The mode used to open the collection copied. Default is 'w'.
+            npartitions: The number of partitions top copy in parallel. Default
+                is number of cores.
+            synchronizer: The synchronizer used to synchronize the collection
+                copied. Default is None.
+
+        Returns:
+            The new collection.
+
+        Example:
+            >>> import zcollection
+            >>> collection = zcollection.open_collection(
+            ...     "my_collection", mode="r")
+            >>> collection.copy(target="my_new_collection")
+        """
+        _LOGGER.info('Copying of the collection to %r', target)
+        if filesystem is None:
+            filesystem = fs_utils.get_fs(target)
+        client = dask_utils.get_client()
+        npartitions = npartitions or dask_utils.dask_workers(client,
+                                                             cores_only=True)
+
+        # Sequence of (source, target) to copy split in npartitions
+        args = tuple(
+            dask_utils.split_sequence(
+                [(item,
+                  fs_utils.join_path(
+                      target,
+                      os.path.relpath(item, self.partition_properties.dir)))
+                 for item in self.partitions(filters=filters)], npartitions))
+        # Copy the selected partitions
+        partial = functools.partial(fs_utils.copy_tree,
+                                    fs_source=self.fs,
+                                    fs_target=filesystem)
+
+        def worker_task(args: Sequence[tuple[str, str]]) -> None:
+            """Function call on each worker to copy the partitions."""
+            tuple(map(lambda arg: partial(*arg), args))
+
+        client.gather(client.map(worker_task, args))
+        # Then the remaining files in the root directory (config, metadata,
+        # etc.)
+        fs_utils.copy_files([
+            item['name']
+            for item in self.fs.listdir(self.partition_properties.dir,
+                                        detail=True) if item['type'] == 'file'
+        ], target, self.fs, filesystem)
+        return Collection.from_config(target,
+                                      mode=mode,
+                                      filesystem=filesystem,
+                                      synchronizer=synchronizer)
